{% extends "base.html" %}

{% block title %}IceProd - Login{% end %}

{% block page_title %}Login{% end %}

{% block body %}

<form action="/login" method="post">
{% module xsrf_form_html() %}
  <input type="hidden" name="next" value="{{ next }}" />
  <div style="margin: 1em 0">
    <div>Username: <input type="text" name="username" /></div>
    <div>Password: <input type="password" name="password" /></div>
    <div style="margin: .2em 0"><input type="submit" value="Sign in" /></div>
    {% if status == 'failed' %}
    <h4>Failed to login</h4>
    {% end %}
  </div>
</form>

{% end %}