#overlay {
    display:none;
    position:fixed;
    left:0px;
    top:0px;
    width:100%;
    height:100%;
    background:#000;
    opacity:0.4;
    z-index:99999;
}

#popup {
    display:none;
    position:fixed;
    left:50%;
    width:600px;
    top:0;
    min-height:600px;
    margin-top:30px;
    margin-left:-300px;
    background:#fff;
    border:2px solid #000;
    z-index:100000;
    padding: 10px;
    box-shadow: 5px 5px 10px #888888;
}
#popup h1, #popup h2, #popup h3 {
    font-weight: bold;
}
#popup p {
    margin-left: 30px;
    margin-bottom: 10px;
}
#popup li {
    margin-left: 15px;
}
a.header.key {
    color: black;
}